"""
Schema
"""
import os
from typing import Any, Optional

import pooch
from pydantic import BaseModel, Field

from bookshelf.utils import get_env_var, get_notebook_directory

Version = str
Edition = int


class BookVersion(BaseModel):
    """
    Version information for a book
    """

    version: Version
    edition: Edition
    url: str
    hash: str
    private: Optional[bool] = False


class VolumeMeta(BaseModel):
    """
    Schema for a given Volume (A collection of Books with the same name)
    """

    name: str
    license: str  # A change in license will require a new volume
    versions: list[BookVersion]

    def get_latest_version(self) -> Version:
        """
        Get the latest version for a volume

        Returns
        -------
        Version string
        """
        ordered_versions = sorted([v.version for v in self.versions if not v.private])
        if not ordered_versions:
            raise ValueError("No published volumes")

        return ordered_versions[-1]

    def get_version(self, version: Version) -> list[BookVersion]:
        """
        Get a set of books for a given version

        Returns
        -------
        List of matching books sorted by edition
        """
        matching_versions = []

        for version_meta in self.versions:
            if version_meta.version == version:
                matching_versions.append(version_meta)

        matching_versions = sorted(matching_versions, key=lambda v: v.edition)
        return matching_versions


class FileDownloadInfo(BaseModel):
    """
    A File to be downloaded as part of a dataset
    """

    url: str
    """
    URL of the file to be downloaded

    This can be any URLs supported by :func:`pooch.download_file`, or a local file if the prefix "file://" is
    used. With local files, the filename must be a path relative to the notebooks directory. For example,
    `file://gdp-ndc-tool/13Mar2023a_CR_gdp_results.csv`.
    """
    hash: str
    """
    Hash of the file to be downloaded

    Used to verify the consistency of the downloaded file
    """


class ControlledVocabularyValue(BaseModel):
    """
    A value in a controlled vocabulary
    """

    value: str
    """
    Value of the controlled vocabulary
    """
    description: str
    """
    Description of the controlled vocabulary value
    """


class Dimension(BaseModel):
    """
    Dimension information

    A dimension describes a feature of the metadata (expressed as a column
    in the index of an {py:class}`scmdata.ScmRun`). This dimension can optionally be controlled by
    a controlled vocabulary which limits the possible values.
    """

    name: str
    """
    Name of the metadata dimension

    This is the same as the column in {py:class}`scmdata.ScmRun`
    """
    description: str
    """
    Description of the metadata dimension
    """
    type: str
    """
    Type of the values in the metadata dimension

    This is not currently verified
    """
    required_column: bool = Field(default=True)
    """
    Indication about whether this metadata dimension is compulsory in all the dataset in the book
    """
    allowed_NA: bool
    """
    Indication about whether all values in this metadata dimension must be non-empty
    """
    controlled_vocabulary: Optional[list[ControlledVocabularyValue]] = None
    """
    List of possible controlled vocabulary of this metadata dimension
    """


class DatasetMetadata(BaseModel):
    """
    Metadata about a dataset

    A dataset may consist of multiple files (:class:`FileDownloadInfo`)
    """

    url: Optional[str]
    doi: Optional[str]
    files: list[FileDownloadInfo] = Field(default_factory=list)
    author: str


class VersionMetadata(BaseModel):
    """
    Metadata about a single version of a book
    """

    version: Version
    dataset: DatasetMetadata
    private: Optional[bool] = Field(default=False)


class NotebookMetadata(BaseModel):
    """
    Schema for a running a notebook

    This represents the metadata for a selected version of a volume
    """

    name: str
    version: Version
    edition: Edition
    description: Optional[str]
    license: str
    source_file: str
    private: bool
    metadata: dict[str, Any]  # TODO: type this
    dataset: DatasetMetadata
    data_dictionary: list[Dimension] = Field(default_factory=list)

    def long_name(self) -> str:
        """
        Long name of the book

        Includes name and long version

        Returns
        -------
        str
            Long identifier for a book
        """
        return f"{self.name}@{self.long_version()}"

    def long_version(self) -> str:
        """
        Long version identifier

        Of the form "{version}_e{edition}" e.g. v1.0.1_e002.

        Returns
        -------
        str
            Version identification string
        """
        return f"{self.version}_e{self.edition:03}"

    def download_file(self, idx: int = 0) -> str:
        """
        Download a dataset file

        Uses ``pooch`` to manage the downloading, verification and caching of data file.
        The first call will trigger a download and subsequent calls may use the cached
        file if the previous download succeeded.

        Parameters
        ----------
        idx
            Index of the file to download (0-based). Defaults to the first file if
            no value is provided

        Returns
        -------
        str
            Filename of the locally downloaded file
        """
        cache_location = get_env_var("DOWNLOAD_CACHE_LOCATION", raise_on_missing=False, default=None)

        try:
            file_info = self.dataset.files[idx]
        except IndexError as e:
            raise ValueError("Requested index does not exist") from e

        file_hash: Optional[str] = file_info.hash
        if not file_hash:
            # replace an empty string with None
            file_hash = None

        if file_info.url.startswith("file://"):
            return os.path.join(get_notebook_directory(), file_info.url[7:])

        res: str = pooch.retrieve(
            file_info.url,
            known_hash=file_hash,
            path=cache_location,
        )
        return res


class ConfigSchema(BaseModel):
    """
    Schema for a given Volume (A collection of Books with the same name)

    A volume can hold multiple versions of the same data
    """

    name: str
    edition: Edition
    description: Optional[str]
    license: str
    source_file: str
    metadata: dict[str, Any]  # TODO: type this
    versions: list[VersionMetadata]
    data_dictionary: list[Dimension] = Field(default_factory=list)



"""
Constants
"""
import os

DATA_FORMAT_VERSION = "v0.2.1"
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))

DATA_DIR = os.path.join(ROOT_DIR, "data")
RAW_DATA_DIR = os.path.join(DATA_DIR, "raw")
TEST_DATA_DIR = os.path.join(ROOT_DIR, "tests", "test-data")
PROCESSED_DATA_DIR = os.path.join(DATA_DIR, "processed", DATA_FORMAT_VERSION)


DEFAULT_S3_BUCKET = "cr-prod-datasets-bookshelf"
"""The default bucket used to store the raw data files"""
DEFAULT_BOOKSHELF = f"https://{DEFAULT_S3_BUCKET}.s3.us-west-2.amazonaws.com/{DATA_FORMAT_VERSION}"
"""Default URL for the remote bookshelf"""
ENV_PREFIX = "BOOKSHELF_"



"""
A collection of curated climate data sets
"""
import importlib.metadata

from bookshelf.book import LocalBook
from bookshelf.shelf import BookShelf

__version__ = importlib.metadata.version("bookshelf")

__all__ = ["LocalBook", "BookShelf"]



"""
A BookShelf is a collection of Books that can be queried and fetched as needed.
"""
import json
import logging
import os
import pathlib
from collections.abc import Iterable
from typing import TYPE_CHECKING, Optional, Union, cast

import boto3
import boto3.exceptions
import datapackage
import requests.exceptions

from bookshelf.book import LocalBook
from bookshelf.constants import DATA_FORMAT_VERSION, DEFAULT_S3_BUCKET
from bookshelf.errors import UnknownBook, UnknownEdition, UnknownVersion, UploadError
from bookshelf.schema import BookVersion, Edition, Version, VolumeMeta
from bookshelf.utils import (
    build_url,
    create_local_cache,
    fetch_file,
    get_env_var,
    get_remote_bookshelf,
)

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client


logger = logging.getLogger(__name__)


def _fetch_volume_meta(
    name: str,
    remote_bookshelf: str,
    local_bookshelf: pathlib.Path,
    force: bool = True,
) -> VolumeMeta:
    """
    Fetch information about the books available for a given volume

    Parameters
    ----------
    name : str
        Name of the volume to fetch
    remote_bookshelf : str
        URL for the remote bookshelf
    local_bookshelf : pathlib.Path
        Local path where downloaded books will be stored.

        Must be a writable directory
    force: bool
        If True metadata is always fetched from the remote bookshelf

    Returns
    -------
    VolumeMeta
    """
    fname = "volume.json"

    local_fname = local_bookshelf / name / fname
    url = build_url(remote_bookshelf, name, fname)

    fetch_file(url, local_fname, force=force)

    with open(str(local_fname)) as file_handle:
        data = json.load(file_handle)

    return VolumeMeta(**data)


def _upload_file(s3: "S3Client", bucket: str, key: str, fname: str) -> None:
    try:
        logger.info(f"Uploading {fname} to {bucket} - {key}")
        s3.upload_file(fname, bucket, key, ExtraArgs={"ACL": "public-read"})
    except boto3.exceptions.S3UploadFailedError as s3_error:
        msg = f"Failed to upload {fname} to s3"
        logger.exception(msg)
        raise UploadError(msg) from s3_error


def _update_volume_meta(book: LocalBook, remote_bookshelf: str) -> str:
    try:
        volume_meta = _fetch_volume_meta(book.name, remote_bookshelf, book.local_bookshelf, force=True)
    except requests.exceptions.HTTPError:
        volume_meta = VolumeMeta(name=book.name, license="", versions=[])

    meta_fname = str(book.local_bookshelf / book.name / "volume.json")
    volume_meta.versions.append(
        BookVersion(
            **{
                "version": book.version,
                "edition": book.edition,
                "url": book.url(),
                "hash": book.hash(),
                "private": book.metadata().get("private", False),
            }
        )
    )
    with open(meta_fname, "w") as file_handle:
        file_handle.write(volume_meta.json())

    return meta_fname


class BookShelf:
    """
    A BookShelf stores a number of Books

    If a Book isn't available locally, it will be queried from the remote bookshelf.

    Books can be fetched using :func:`load` by name. Specific versions of a book can be
    pinned if needed otherwise the latest version of the book is loaded.
    """

    def __init__(
        self,
        path: Union[str, pathlib.Path, None] = None,
        remote_bookshelf: Optional[str] = None,
    ):
        if path is None:
            path = create_local_cache(path)
        self.path = pathlib.Path(path)
        self.remote_bookshelf = get_remote_bookshelf(remote_bookshelf)

    def load(
        self,
        name: str,
        version: Optional[Version] = None,
        edition: Optional[Edition] = None,
        force: bool = False,
    ) -> LocalBook:
        """
        Load a book

        If the book's metadata does not exist locally or an unknown version is requested
        the remote bookshelf is queried, otherwise the local metadata is used.

        Parameters
        ----------
        name: str
            Name of the volume to load
        version: str
            Version to load

            If no version is provided, the latest version is returned
        edition: int
            Edition of book to load

            If no edition is provided, the latest edition of the selected version is returned
        force: bool
            If True, redownload the book metadata

        Raises
        ------
        UnknownVersion
            The requested version is not available for the selected volume
        UnknownBook
            An invalid volume is requested

        Returns
        -------
        :class:`LocalBook`
            A book from which the resources can be accessed
        """
        if version is None or edition is None or force:
            version, edition = self._resolve_version(name, version, edition)
        metadata_fragment = LocalBook.relative_path(name, version, edition, "datapackage.json")
        metadata_fname = self.path / metadata_fragment
        if not metadata_fname.exists():
            try:
                url = build_url(
                    self.remote_bookshelf,
                    *LocalBook.path_parts(name, version, edition, "datapackage.json"),
                )
                fetch_file(
                    url,
                    local_fname=metadata_fname,
                    known_hash=None,
                    force=force,
                )
            except requests.exceptions.HTTPError as http_error:
                raise UnknownVersion(name, version) from http_error

        if not metadata_fname.exists():
            raise AssertionError()
        return LocalBook(name, version, edition, local_bookshelf=self.path)

    def is_available(
        self,
        name: str,
        version: Optional[Version] = None,
        edition: Optional[Edition] = None,
    ) -> bool:
        """
        Check if a Book is available from the remote bookshelf

        Parameters
        ----------
        name : str
            Name of the volume to check
        version : str
            Version of the volume to check

            If no version is provided, then check if any Book's with a matching name
            have been uploaded.

        Returns
        -------
        bool
            True if a Book with a matching name and version exists on the remote bookshelf
        """
        try:
            self._resolve_version(name, version, edition)
        except (UnknownBook, UnknownVersion, UnknownEdition):
            return False
        return True

    def is_cached(self, name: str, version: Version, edition: Edition) -> bool:
        """
        Check if a book with a matching name/version is cached on the local bookshelf

        Parameters
        ----------
        name : str
            Name of the volume to check
        version : str
            Version of the volume to check
        edition : int
            Edition of the volume to check

        Returns
        -------
        bool
            True if a matching book is cached locally
        """
        try:
            # Check if the metadata for the book can be successfully read
            book = LocalBook(name, version, edition, local_bookshelf=self.path)
            book.metadata()
        except FileNotFoundError:
            return False
        return True

    def publish(self, book: LocalBook, force: bool = False) -> None:
        """
        Publish a book to the remote bookshelf

        Parameters
        ----------
        book : LocalBook
            Book to upload
        force : bool
            If True, overwrite any existing data

        Raises
        ------
        UploadError
            Unable to upload the book to the remote bookshelf. See error message for
            more information about how to resolve this issue.
        """
        if self.is_available(book.name, book.version):
            remote_book = self.load(book.name, book.version)

            if remote_book.edition >= book.edition:
                msg = (
                    "Edition value has not been increased (remote:"
                    f" {remote_book.long_version()}, local: {book.long_version()})"
                )
                if not force:
                    raise UploadError(msg)
                logger.error(msg)
            logger.warning("Uploading a new edition of an existing book")
        files = book.files()

        # Check if additional files are going to be uploaded
        resources = book.as_datapackage().resources
        resource_fnames = [
            resource.descriptor["filename"] for resource in cast(Iterable[datapackage.Resource], resources)
        ]
        for resource_file in files:
            fname = os.path.basename(resource_file)
            if fname == "datapackage.json":
                continue
            if fname not in resource_fnames:
                raise UploadError(f"Non-resource file {fname} found in book")

        # Upload using boto3 by default for testing
        # Maybe support other upload methods in future

        s3 = boto3.client("s3")
        bucket = get_env_var("BUCKET", add_prefix=True, default=DEFAULT_S3_BUCKET)
        prefix = get_env_var("BUCKET_PREFIX", add_prefix=True, default=DATA_FORMAT_VERSION)

        logger.info(f"Beginning to upload {book.name}@{book.version}")
        for resource_file in files:
            key = "/".join(
                (
                    prefix,
                    book.name,
                    book.long_version(),
                    os.path.basename(resource_file),
                )
            )
            _upload_file(s3, bucket, key, resource_file)

        # Update the metadata with the latest version information
        # Note that this doesn't have any guardrails and is susceptible to race conditions
        # Shouldn't be a problem for testing, but shouldn't be used in production
        meta_fname = _update_volume_meta(book, self.remote_bookshelf)
        key = "/".join((prefix, book.name, os.path.basename(meta_fname)))
        _upload_file(s3, bucket, key, meta_fname)

        logger.info(f"Book {book.name}@{book.version} ed.{book.edition} uploaded successfully")

    def _resolve_version(
        self,
        name: str,
        version: Optional[Version] = None,
        edition: Optional[Edition] = None,
    ) -> tuple[Version, Edition]:
        # Update the package metadata
        try:
            meta = _fetch_volume_meta(name, self.remote_bookshelf, self.path)
        except requests.exceptions.HTTPError as http_error:
            raise UnknownBook(f"No metadata for {name!r}") from http_error

        if version is None:
            version = meta.get_latest_version()

        # Verify that the version exists
        matching_version_books = meta.get_version(version)
        if not matching_version_books:
            raise UnknownVersion(name, version)

        # Find edition
        if edition is None:
            edition = matching_version_books[-1].edition
        if edition not in [b.edition for b in matching_version_books]:
            raise UnknownEdition(name, version, edition)
        return version, edition

    def list_versions(self, name: str) -> list[str]:
        """
        Get a list of available versions for a given Book

        Parameters
        ----------
        name: str
            Name of book

        Returns
        -------
        list of str
            List of available versions
        """
        try:
            meta = _fetch_volume_meta(name, self.remote_bookshelf, self.path)
        except requests.exceptions.HTTPError as http_error:
            raise UnknownBook(f"No metadata for {name!r}") from http_error

        return [version.version for version in meta.versions if not version.private]

    def list_books(self) -> list[str]:
        """
        Get a list of book names

        Returns
        -------
        list of str
            List of available books
        """
        raise NotImplementedError



"""
Dataset structure

This module offers some functions to display the structure of a dataset
"""


from __future__ import annotations

from collections.abc import Iterable

from scmdata import ScmRun


def get_dataset_dictionary(data: ScmRun) -> dict[str, Iterable[str | float | int]]:
    """
    Extract unique metadata values from an ScmRun object.

    This function iterates through the columns of the metadata in the ScmRun object
    and creates a dictionary. Each key in the dictionary corresponds to a column name,
    and the associated value is a list of unique values in that column.

    Parameters
    ----------
    data
        The input ScmRun object from which metadata will be extracted.

    Returns
    -------
        A dictionary with column names as keys and lists of unique values in those
        columns as values. The values are extracted from the metadata of the ScmRun object.
    """
    data_dict: dict[str, Iterable[str | int | float]] = {}
    for column in data.meta_attributes:
        data_dict[column] = data.get_unique_meta(column)

    return data_dict


def print_dataset_structure(data: ScmRun) -> None:
    """
    Print the structure of a dataset.

    This function displays a tabular representation of the dataset's dimensions
    and their unique values. Each row corresponds to a unique value in a dimension,
    and each column represents a different dimension of the dataset.

    Parameters
    ----------
    data
        The input dataset in the form of an ScmRun object.

    Returns
    -------
        This function does not return anything. It prints the dataset structure
        directly to the console.
    """
    data_dict = get_dataset_dictionary(data)

    # Convert values to strings once
    k_lst = list(data_dict.keys())
    v_lst = [list(map(str, values)) for values in data_dict.values()]
    max_length = max(len(values) for values in v_lst)

    # Calculate width for each column
    width_lst = [max(len(str(item)) for item in [key, *values]) + 5 for key, values in zip(k_lst, v_lst)]

    # Print header
    print("".join(f"{item:{width}}" for item, width in zip(k_lst, width_lst)))

    # Print header divider
    print(
        "".join(
            f"{item:{width}}" for item, width in zip(["-" * (width - 5) for width in width_lst], width_lst)
        )
    )

    # Print each row of values
    for i in range(max_length):
        row_values = [values[i] if i < len(values) else "" for values in v_lst]
        print("".join(f"{item:{width}}" for item, width in zip(row_values, width_lst)))
